@@ -36,6 +36,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
+#include <limits.h>
 
 #define _GNU_SOURCE
 #include <getopt.h>
@@ -56,6 +57,7 @@
 #define KEY_STATUS  '\x16' /* C-v: show program option */
 #define KEY_SEND    '\x13' /* C-s: send file */
 #define KEY_RECEIVE '\x12' /* C-r: receive file */
+#define KEY_BREAK   '\x1c' /* C-\: break */
 
 #define STO STDOUT_FILENO
 #define STI STDIN_FILENO
@@ -72,6 +74,9 @@ struct {
 	int databits;
 	int noinit;
 	int noreset;
+#ifdef UUCP_LOCK_DIR
+	int nolock;
+#endif
 	unsigned char escape;
 	char send_cmd[128];
 	char receive_cmd[128];
@@ -85,6 +90,9 @@ struct {
 	.databits = 8,
 	.noinit = 0,
 	.noreset = 0,
+#ifdef UUCP_LOCK_DIR
+	.nolock = 0,
+#endif
 	.escape = '\x01',
 	.send_cmd = "ascii_xfr -s -v -l10",
 	.receive_cmd = "rz -vv"
@@ -94,6 +102,88 @@ int tty_fd;
 
 /**********************************************************************/
 
+#ifdef UUCP_LOCK_DIR
+
+/* use HDB UUCP locks  .. see
+ * <http://www.faqs.org/faqs/uucp-internals> for details
+ */
+
+char lockname[_POSIX_PATH_MAX] = "";
+
+int
+uucp_lockname(const char *dir, const char *file)
+{
+	char *p, *cp;
+	struct stat sb;
+
+	if ( ! dir || *dir == '\0' || stat(dir, &sb) != 0 )
+		return -1;
+
+	/* cut-off initial "/dev/" from file-name */
+	p = strchr(file + 1, '/');
+	p = p ? p + 1 : (char *)file;
+	/* replace '/'s with '_'s in what remains (after making a copy) */
+	p = cp = strdup(p);
+	do { if ( *p == '/' ) *p = '_'; } while(*p++);
+	/* build lockname */
+	snprintf(lockname, sizeof(lockname), "%s/LCK..%s", dir, cp);
+	/* destroy the copy */
+	free(cp);
+
+	return 0;
+}
+
+int
+uucp_lock(void)
+{
+	int r, fd, pid;
+	char buf[16];
+	mode_t m;
+
+	if ( lockname[0] == '\0' ) return 0;
+
+	fd = open(lockname, O_RDONLY);
+	if ( fd >= 0 ) {
+		r = read(fd, buf, sizeof(buf)); 
+		close(fd);
+		/* if r == 4, lock file is binary (old-style) */
+		pid = (r == 4) ? *(int *)buf : strtol(buf, NULL, 10);
+		if ( pid > 0 
+			 && kill((pid_t)pid, 0) < 0 
+			 && errno == ESRCH ) {
+			/* stale lock file */
+			printf("Removing stale lock: %s\n", lockname);
+			sleep(1);
+			unlink(lockname);
+		} else {
+			lockname[0] = '\0';
+			errno = EEXIST;
+			return -1;
+		}
+	}
+	/* lock it */
+	m = umask(022);
+	fd = open(lockname, O_WRONLY|O_CREAT|O_EXCL, 0666);
+	if ( fd < 0 ) { lockname[0] = '\0'; return -1; }
+	umask(m);
+	snprintf(buf, sizeof(buf), "%04d\n", getpid());
+	write(fd, buf, strlen(buf));
+	close(fd);
+
+	return 0;
+}
+
+int
+uucp_unlock(void)
+{
+	if ( lockname[0] ) unlink(lockname);
+	return 0;
+}
+
+#endif /* of UUCP_LOCK_DIR */
+
+/**********************************************************************/
+
 ssize_t
 writen_ni(int fd, const void *buff, size_t n)
 {
@@ -154,6 +244,10 @@ fatal (const char *format, ...)
 	/* wait a bit for output to drain */
 	sleep(1);
 
+#ifdef UUCP_LOCK_DIR
+	uucp_unlock();
+#endif
+	
 	exit(EXIT_FAILURE);
 }
 
@@ -294,36 +388,46 @@ bits_next (int bits)
 /**********************************************************************/
 
 void
-child_int_handler (int signum)
+child_empty_handler (int signum)
 {
 }
 
 void
 establish_child_signal_handlers (void)
 {
-	struct sigaction int_action;
+	struct sigaction empty_action;
 	
-	/* Set up the structure to specify the exit action. */
-    int_action.sa_handler = child_int_handler;
-	sigemptyset (&int_action.sa_mask);
-	int_action.sa_flags = 0;
+	/* Set up the structure to specify the "empty" action. */
+    empty_action.sa_handler = child_empty_handler;
+	sigemptyset (&empty_action.sa_mask);
+	empty_action.sa_flags = 0;
 	
-	sigaction (SIGINT, &int_action, NULL);
+	sigaction (SIGINT, &empty_action, NULL);
+	sigaction (SIGTERM, &empty_action, NULL);
 }
 
 int
 run_cmd(int fd, ...)
 {
 	pid_t pid;
+	sigset_t sigm, sigm_old;
+
+	/* block signals, let child establish its own handlers */
+	sigemptyset(&sigm);
+	sigaddset(&sigm, SIGTERM);
+	sigprocmask(SIG_BLOCK, &sigm, &sigm_old);
 
 	pid = fork();
 	if ( pid < 0 ) {
+		sigprocmask(SIG_SETMASK, &sigm_old, NULL);
 		fd_printf(STO, "*** cannot fork: %s\n", strerror(errno));
 		return -1;
 	} else if ( pid ) {
 		/* father: picocom */
 		int r;
 
+		/* reset the mask */
+		sigprocmask(SIG_SETMASK, &sigm_old, NULL);
 		/* wait for child to finish */
 		waitpid(pid, &r, 0);
 		/* reset terminal (back to raw mode) */
@@ -344,6 +448,7 @@ run_cmd(int fd, ...)
 		char cmd[512];
 
 		establish_child_signal_handlers();
+		sigprocmask(SIG_SETMASK, &sigm_old, NULL);
 		/* unmanage terminal, and reset it to canonical mode */
 		term_remove(STI);
 		/* unmanage serial port fd, without reset */
@@ -441,7 +546,12 @@ loop(void)
 
 			case ST_COMMAND:
 				if ( c == opts.escape ) {
-					state = ST_COMMAND;
+					state = ST_TRANSPARENT;
+					/* pass the escape character down */
+					if (tty_q.len <= TTY_Q_SZ)
+						tty_q.buff[tty_q.len++] = c;
+					else
+						fd_printf(STO, "\x07");
 					break;
 				}
 				state = ST_TRANSPARENT;
@@ -540,6 +650,10 @@ loop(void)
 					else
 						run_cmd(tty_fd, opts.receive_cmd, NULL);
 					break;
+				case KEY_BREAK:
+					term_break(tty_fd);
+					fd_printf(STO, "\r\n*** break sent ***\r\n");
+					break;
 				default:
 					break;
 				}
@@ -605,6 +719,9 @@ deadly_handler(int signum)
 {
 	kill(0, SIGTERM);
 	sleep(1);
+#ifdef UUCP_LOCK_DIR
+	uucp_unlock();
+#endif
 	exit(EXIT_FAILURE);
 }
 
@@ -653,6 +770,7 @@ show_usage(char *name)
 	printf("  --<e>scape <char>\n");
 	printf("  --no<i>nit\n");
 	printf("  --no<r>eset\n");
+	printf("  --no<l>ock\n");
 	printf("  --<s>end-cmd <command>\n");
 	printf("  --recei<v>e-cmd <command>\n");
 	printf("  --<h>elp\n");
@@ -672,6 +790,7 @@ parse_args(int argc, char *argv[])
 		{"escape", required_argument, 0, 'e'},
 		{"noinit", no_argument, 0, 'i'},
 		{"noreset", no_argument, 0, 'r'},
+		{"nolock", no_argument, 0, 'l'},
 		{"flow", required_argument, 0, 'f'},
 		{"baud", required_argument, 0, 'b'},
 		{"parity", required_argument, 0, 'p'},
@@ -687,7 +806,7 @@ parse_args(int argc, char *argv[])
 		/* no default error messages printed. */
 		opterr = 0;
 
-		c = getopt_long(argc, argv, "hirs:r:e:f:b:p:d:",
+		c = getopt_long(argc, argv, "hirls:r:e:f:b:p:d:",
 						longOptions, &optionIndex);
 
 		if (c < 0)
@@ -708,6 +827,9 @@ parse_args(int argc, char *argv[])
 		case 'r':
 			opts.noreset = 1;
 			break;
+		case 'l':
+			opts.nolock = 1;
+			break;
 		case 'e':
 			if ( isupper(optarg[0]) )
 				opts.escape = optarg[0] - 'A' + 1;
@@ -808,6 +930,7 @@ parse_args(int argc, char *argv[])
 	printf("escape is      : C-%c\n", 'a' + opts.escape - 1);
 	printf("noinit is      : %s\n", opts.noinit ? "yes" : "no");
 	printf("noreset is     : %s\n", opts.noreset ? "yes" : "no");
+	printf("nolock is      : %s\n", opts.nolock ? "yes" : "no");
 	printf("send_cmd is    : %s\n", opts.send_cmd);
 	printf("receive_cmd is : %s\n", opts.receive_cmd);
 	printf("\n");
@@ -828,6 +951,12 @@ main(int argc, char *argv[])
 	if ( r < 0 )
 		fatal("term_init failed: %s", term_strerror(term_errno, errno));
 
+#ifdef UUCP_LOCK_DIR
+	if ( ! opts.nolock ) uucp_lockname(UUCP_LOCK_DIR, opts.port);
+	if ( uucp_lock() < 0 )
+		fatal("cannot lock %s: %s", opts.port, strerror(errno));
+#endif
+
 	tty_fd = open(opts.port, O_RDWR | O_NONBLOCK);
 	if (tty_fd < 0)
 		fatal("cannot open %s: %s", opts.port, strerror(errno));
@@ -875,6 +1004,10 @@ main(int argc, char *argv[])
 	/* wait a bit for output to drain */
 	sleep(1);
 
+#ifdef UUCP_LOCK_DIR
+	uucp_unlock();
+#endif
+
 	return EXIT_SUCCESS;
 }
 
